"""
Script used to collect SPP logs from the command line. This script also augments 
the collected logs with additional logs (such as /var/log/messages) if executed 
with the proper permissions. In order to include all logs in the collection 
process, it is suggested that this script be executed using sudo.

:author Justin Albano <jalbano@catalogicsoftware.com>:
:since 10.1.8:
"""

import requests
import getpass
import json
import sys
import argparse
import traceback
import os
import re
import time
import zipfile
import textwrap

from requests import ConnectionError


# The directory that this script resides in.
CWD = os.path.dirname(os.path.realpath(__file__))

# The maximum width of the output displayed on the command line.
LINE_WIDTH = 80


# ==============================================================================
# Formatting
# ==============================================================================

def print_step(step, indent_width=3):

    step_len = len(step)

    result_width = 7
    whitespace_width = 2
    max_width = LINE_WIDTH - indent_width - result_width - whitespace_width

    if step_len < max_width:
        dots = '.' * (max_width - step_len)
    else:
        dots = '...'

    indent = ' ' * indent_width
    sys.stdout.write('{}{} {} '.format(indent, step, dots))
    sys.stdout.flush()


def print_error(e):
    print('')
    print('-' * LINE_WIDTH)
    print(' Error')
    print('-' * LINE_WIDTH)
    print(' {}'.format(str(e)))
    print('')
    traceback.print_exc()
    print('-' * LINE_WIDTH)


def print_header():

    print('')
    print('=' * LINE_WIDTH)
    print(' SPP Log Collection Tool')
    print('=' * LINE_WIDTH)
    print('')


def print_footer(completion_time_ms, destination):

    print('')
    print('-' * LINE_WIDTH)
    print(' Collection Completed')
    print('-' * LINE_WIDTH)
    print('')
    print('   Elapsed time: {}'.format(format_time(completion_time_ms)))
    print('   Location:     {}'.format(destination))
    print('')
    print('-' * LINE_WIDTH)
    print('')


def current_time_ms():
    # https://stackoverflow.com/a/5998359/2403253
    return round(time.time() * 1000)


def format_time(millis):
    # https://stackoverflow.com/a/35990338/2403253

    seconds = (millis / 1000) % 60
    seconds = int(seconds)
    minutes = (millis / (1000 * 60)) % 60
    minutes = int(minutes)
    hours = (millis / (1000 * 60* 60)) % 24

    return "%02d:%02d:%02d" % (hours, minutes, seconds)


# ==============================================================================
# Steps
# ==============================================================================

class Context(object):
    """
    Abstraction of the context when executing steps.
    """

    def __init__(self):
        self._last_step_skipped = False

    def skip(self, message=None):
        """
        Skips the current step.

        :param message:
        A reason message for skipping the current step.
        """

        if message is None:
            print('skipped')
        else:
            max_width = LINE_WIDTH - 5
            wrapped_message = textwrap.wrap(message, width=max_width)

            print('skipped')
            print('')

            for line in wrapped_message:
                print('     {}'.format(line))

            print('')

        self._last_step_skipped = True

    @property
    def step_was_skipped(self):
        return self._last_step_skipped

    def reset_for_next_step(self):
        self._last_step_skipped = False
    


def step(context, name, action):
    """
    Executes a step.

    :param name:
    The name of the step.
    :param action:
    The step to the be executed. This step should follow the format:

      lambda context: <step_logic>

    :return:
    The result of the supplied step action.
    """

    try:
        print_step(name)
        result = action(context)

        if not context.step_was_skipped:
            print('success')

        return result
    except:
        print('failure')
        raise
    finally:
        context.reset_for_next_step()


def collect_logs(username, password, download_dir):
    """
    Executes the steps required to collect the logs.

    :param username:
    The username of the user used to connect to SPP.
    :param password:
    The password of the user used to connect to SPP.
    :param download_dir:
    The directory into which the logs should be downloaded.
    """
    
    context = Context()

    print(' Collecting logs')
    print(' ---------------')
    print('')

    session_id = step(
        context,
        'Obtaining session ID',
        lambda context: find_session_id(username, password)
    )
    
    download_response = step(
        context,
        'Downloading logs', 
        lambda context: requests.get(
            'http://localhost:8082/api/endeavour/log/download/diagnostics',
            headers={'x-endeavour-sessionid': session_id}
        )
    )

    destination = step(
        context,
        'Writing logs to file system',
        lambda context: write_zip_to_file_system(download_response, download_dir)
    )

    step(
        context,
        'Collect /var/log/messages',
        lambda context: append_var_log_messages(context, destination)
    )

    return destination


def find_session_id(username, password):
    """
    Finds an SPP session ID corresponding to the supplied username and password.

    :param username:
    The username of the user to create the session ID for.
    :param password:
    The password of the user to create the session ID for.

    :return:
    A session ID corresponding to the supplied username and password.
    """

    try:
        url = 'http://localhost:8082/api/endeavour/session?screenInfo=1'
        headers = {
            'Accept': 'application/json',
            'Content-Type': 'application/json'
        }

        response = requests.post(url, headers=headers, auth=(username, password))
        response_body = json.loads(response.text)
    except ConnectionError:
        raise SppNotRunningException('SPP instance cannot be found...ensure SPP is running and is reachable before running log collector.')

    if 'sessionid' in response_body:
        return response_body['sessionid']
    else:
        raise MissingCredentialsException('Could not obtain session ID from SPP')


def write_zip_to_file_system(download_response, download_dir, chunk_size=128):
    """
    Writes an HTTP response containing a .zip file to the local file system.

    :param download_response: 
    The HTTP response that contains the name of the .zip file in the 
    Content-Disposition header and the content of the .zip file in the response 
    body.
    :param download_dir: 
    The directory to download the .zip into. This is only the directory name and 
    should not contain a file name, as the file name will come from the 
    Content-Disposition header.
    :param chunk_size: 
    The chunk size to use when writing the file.

    :return:
    The file name (full path) of the .zip file written to the local file system.
    """

    zip_name = extract_destination_filename(download_response)
    destination = os.path.join(download_dir, zip_name)

    with open(destination, 'wb') as fd:
        for chunk in download_response.iter_content(chunk_size=chunk_size):
            fd.write(chunk)

    return destination


def extract_destination_filename(response):
    """
    Extracts the intended file name of a .zip file from the Content-Disposition 
    header of the supplied response.

    :param response:
    The HTTP response to extract the file name from.

    :return:
    The file name extracted from the Content-Disposition header.
    """

    disposition = response.headers['content-disposition']
    return re.findall("filename=(.+)", disposition)[0]


def append_var_log_messages(context, zip_file):
    """
    Appends the /var/log/messages file to the supplied .zip file if this script 
    is executed as root (i.e., using sudo).

    :param context:
    The context of the steps being executed.
    :param zip_file:
    The .zip file to write the /var/log/messages file to.
    """

    if not is_root():
        context.skip('Collection of /var/log/messages can only be completed with \
            administrative access. To collect /var/log/messages, please execute \
            this script again as sudo (sudo <script>).')
    else:
        target = 'system/var/log/messages'

        with zipfile.ZipFile(zip_file, 'a', allowZip64=True) as f:

            zip_files = f.namelist()

            if len(zip_files) > 0:

                shortest_file_path = min(zip_files, key=len)
                shortest_file_path_parts = shortest_file_path.split('/')
                top_level_dir = shortest_file_path_parts[0]

                target = os.path.join(top_level_dir, target)

            f.write('/var/log/messages', target)


def is_root():
    """
    Checks if this script is being executed as root (i.e., sudo).

    :return:
    True if executed as root; False otherwise.
    """
    return os.geteuid() == 0


# ==============================================================================
# Exceptions & Errors
# ==============================================================================


class MissingCredentialsException(Exception):
    
    def __init__(self, message):
        Exception.__init__(self, message)


class SppNotRunningException(Exception):
    
    def __init__(self, message):
        Exception.__init__(self, message)


# ==============================================================================
# Script Execution
# ==============================================================================

def get_credentials(args):
    """
    Extracts the credentials from the supplied arguments. If username only is 
    not provided in the supplied arguments, the user will be prompted for them. 
    If the password only is not supplied in the arguments, the user will be 
    prompted for them. If both username and password are not supplied in the 
    arguments, both will be prompted from the user, but if both are supplied in 
    the arguments, no prompt will be displayed.

    :return:
    The extracted username and password in a tuple: (username, password).
    """

    username = args.username
    password = args.password

    if username is None or password is None:
        
        print('  Enter SPP credentials')
        print('')

        if username is None:
            username = raw_input('    Username: ')

        if password is None:
            password = getpass.getpass('    Password: ')

        print('')

    return (username, password)



if __name__ == "__main__":
    
    try:
        parser = argparse.ArgumentParser(description='SPP Log Collection Tool')
        parser.add_argument('-u', '--username', type=str, help='Username used to authenticate with SPP')
        parser.add_argument('-p', '--password', type=str, help='Password used to authenticate with SPP')
        parser.add_argument('-d', '--destination', type=str, help='The directory to write the logs to')
        args = parser.parse_args()

        log_destination_dir = args.destination if args.destination is not None else CWD

        print_header()

        start = current_time_ms()

        (username, password) = get_credentials(args)
        destination = collect_logs(username, password, log_destination_dir)

        end = current_time_ms();
        completion_time_ms = end - start

        print_footer(completion_time_ms, destination)

    except Exception as e:
        print_error(e)

