#!/usr/bin/python

####################################################################################################################
##### Script installs/updates the HTTPS (or) LDAP certifcate
#####
##### Syntax: certificate-manager.pyc <certificate type> <deployment type> <temporary directory where file exists>
##### Example: certificate-manager.pyc https deploy /tmp/temporarydirectory
##### Example: certificate-manager.pyc ldap deploy /tmp/temporarydirectory
##### 
##### External service is expected to copy the new Certificate file under temporary directory  prior to invocation
####################################################################################################################

import subprocess
import string
import random
import os, shutil, logging, sys
from optparse import OptionParser

logger = None
CONST = None
FORWARD_SLASH="/"

change_https_certificate_file_name = False
uploaded_https_certificate_found = False

uploaded_ldap_certificate_found = False

################################################################################
##### MODIFY THIS SECTION FOR LOCATION/CONFIGURATION CHANGES FOR HTTPS
################################################################################
class HTTPSCONST(object):
	SSL_CERTIFICATE_BACKUP_DIR="/tmp/spp-httpsbackup/"	
	SSL_CERTIFICATE_DIR="/opt/virgo/configuration/"
	SSL_CERTIFICATE_FILE_NAME="ecx-beta.crt"
	SSL_CERTIFICATE_KEY_FILE_NAME="ecx-beta.key"
	SSL_CERTIFICATE_FILE="/opt/virgo/configuration/ecx-beta.crt"
	SSL_CERTIFICATE_KEY_FILE="/opt/virgo/configuration/ecx-beta.key"

	def __setattr__(self, *_):
		pass

################################################################################
##### MODIFY THIS SECTION FOR LOCATION/CONFIGURATION CHANGES FOR LDAP
################################################################################
class LDAPCONST(object):	

	def __setattr__(self, *_):
		pass

################################################################################

def main(argv):

	global logger    
	logger = logging.getLogger(__name__)
	logger.setLevel(logging.INFO)

	# create console handler and set level to info
	console = logging.StreamHandler()
	console.setLevel(logging.ERROR)
	consoleformatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
	console.setFormatter(consoleformatter)
	logger.addHandler(console)

	# create a file handler
	handler = logging.FileHandler(os.path.join('/data/log/adminconsole','certificate-manager.log'))
	handler.setLevel(logging.INFO)
	# create a logging format
	formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
	handler.setFormatter(formatter)
	# add the handlers to the logger
	logger.addHandler(handler)
    
	if len(sys.argv) == 5:
		add_wide_gaps()
		logger.info('##########################################################################################################')
		logger.info('RUNNING CERTIFICATE MANAGER')
		logger.info('Argument Length ' + str(len(argv)))
		logger.info('Arguments: ' + str(argv))
		process(argv[1],argv[2],argv[3],argv[4])
	else:
		logger.error('Invalid number of arguments')
		logger.error('Argument Length ' + str(len(argv)))
		logger.error('Arguments: ' + str(argv))
		logger.error('Syntax:  certificate-manager.pyc <mode> <type> <certfilename> <temporarydirectory>')
		logger.error('Example: certificate-manager.pyc deploy https example.crt /tmp/temporarydirectory')
		logger.error('Example: certificate-manager.pyc deploy ldap example.crt /tmp/temporarydirectory')
		failed_exit()

#################################################
# Create directory (cleans up existing directory)
#################################################
def create_directory(path):
	if (os.path.isdir(path)):
		logger.info('Found existing directory ' + path)
		try:
			logger.info('Removing directory ' + path)
			shutil.rmtree(path, ignore_errors=True)
		except OSError:
			logger.error('Error: %s' % e)
			if e.errno != errno.ENOENT:
			    raise
			    failed_exit()
	
	try:
		logger.info('Creating directory ' + path)
		os.makedirs(path)
	except OSError:
		logger.error('Error: %s' % e)
		if e.errno != errno.ENOENT:
		    raise
		    failed_exit()

#############
# Delete file
#############
def delete_file(file_path):
	if (os.path.isfile(file_path)):
		try:
			logger.info('Deleting ' + file_path)
			os.remove(file_path)
		except OSError as e:
			logger.error('Error: %s' % e)
			if e.errno != errno.ENOENT:
				raise
				failed_exit()

#############
# Copy file
#############
def copy_file(src, dest):
	try:
		logger.info('Copying ' + src + ' to ' + dest)
		shutil.copy2(src, dest)
	except shutil.Error as e:
		logger.error('Error: %s' % e)
		failed_exit()
	except IOError as e:
		logger.error('Error: %s' % e.strerror)
		failed_exit()

#############
# Move file
#############
def move_file(src, dest):
	try:
		logger.info('Moving ' + src + ' to ' + dest)
		shutil.move(src, dest)
	except shutil.Error as e:
		logger.error('Error: %s' % e)
		failed_exit()
	except IOError as e:
		logger.error('Error: %s' % e.strerror)
		failed_exit()

##################
# Delete Directory
##################
def delete_directory(directory):
    try:
        logger.info('Deleting directory ' + directory)
        shutil.rmtree(directory, ignore_errors=True)
    except shutil.Error as e:
        logger.error('Error: %s' % e)
        failed_exit()
    except IOError as e:
        logger.error('Error: %s' % e.strerror)
        failed_exit()


#########################
# Verify Certificate file 
#########################
def verify_file(certificatefile, certtype, temporary_directory):

	logger.info('Verifying certificate mime type ' + certtype)
	cmd = '/usr/bin/file -ib ' + certificatefile 
	p = subprocess.Popen(cmd , shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
	stdout, stderr = p.communicate()

	if p.returncode == 0: 
		
		if certtype == "https":
			mime_type = "text/plain; charset=us-ascii"
		elif certtype == "ldap":
			mime_type = "application/octet-stream; charset=binary"
		else:
			logger.error('Invalid certificate type')
			failed_exit()

		if str(stdout).rstrip() != mime_type:
				logger.info('Invalid certificate mime type ' + str(stdout))
                		delete_directory(temporary_directory)
				failed_exit()
	else:
		logger.error('Failure in verifying certificate file mime type')
		delete_directory(temporary_directory)
		failed_exit()

	logger.info('Certificate file has been verified')


################################################
# Verifies if valid PEM certificate
# Verifies if it contains Key and Certificate
################################################
def verify_valid_pem_certificate(certificatefile):
	
	logger.info('Verifying validity of PEM based certificate container ' + certificatefile)

	cmd = '/usr/bin/openssl x509 -in ' + certificatefile 
	p = subprocess.Popen(cmd , shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
	stdout, stderr = p.communicate()

	if p.returncode != 0:
		logger.error('Certificate section not found ' + certificatefile)
		failed_exit()

	cmd = '/usr/bin/openssl rsa -in ' + certificatefile 
	p = subprocess.Popen(cmd , shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
	stdout, stderr = p.communicate()

	if p.returncode != 0:
		logger.error('Key section not found ' + certificatefile)
		failed_exit()

	logger.info('Valid PEM certicate container')

#############################################
# Extracts private key from certificate file
# For SSL Certificates
#############################################
def extract_private_key():
	logger.info('Extracting private key')
	out = open(CONST.SSL_CERTIFICATE_KEY_FILE,'wb')
	key_seen=False
	with open(CONST.SSL_CERTIFICATE_FILE) as f:
		for line in f:
			if line.startswith("-----BEGIN RSA PRIVATE KEY-----") or line.startswith("-----BEGIN PRIVATE KEY-----"):
				key_seen=True
				out.write(line)
				continue

			if line.startswith("-----END RSA PRIVATE KEY-----") or line.startswith("-----END PRIVATE KEY-----"):
				key_seen=False
				out.write(line)
				break

			if key_seen:
				out.write(line)

#################################################################
# Back end code that will be executed to return data to the user
# For LDAP Certificates
#################################################################
def process_ldap_cert(temporarydirectory, certificatefilename):

	alias=''.join(random.choice(string.ascii_uppercase) for x in range(7))

	LDAP_CERTIFICATE = os.path.join(temporarydirectory,certificatefilename)

	command = '/bin/keytool -import --noprompt -alias ' + alias + ' -keystore /opt/virgo/configuration/keystore -storepass ecx-beta -file ' + LDAP_CERTIFICATE

	logger.info('Command: ' + command)

	pobj = subprocess.Popen(command, bufsize=1, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, cwd='/tmp')

	out, err = pobj.communicate()

	logger.info('Command Returncode : ' + str(pobj.returncode))

	if err:
		logger.error('')
		logger.error('Command Errors :')
		logger.error('----------------')
		logger.error('')
		logger.error(err)

	#delete_directory(temporarydirectory)

	if pobj.returncode == 0:
		if out:
			logger.info('')
			logger.info('Command Output :')
			logger.info('----------------')
			logger.info('')
			logger.info(out)

		clean_exit()
	else:
		if out:
			logger.error('')
			logger.error('Command Output :')
			logger.error('----------------')
			logger.error('')
			logger.error(out)

		failed_exit()

def add_wide_gaps():
        for num in range(1,5): 
            logger.info('')

def clean_exit():
        logger.info('')
        logger.info('Script completed successfully, Exiting!')
        logger.info('')
        sys.exit(0)

def failed_exit():
        logger.info('')
        logger.error('Script completed with failures')        
        sys.exit(1)

def process(mode, deploytype, certificatefilename, temporarydirectory):	
	if mode == "deploy":
		global CONST
		if deploytype == "https":

			logger.info('Processing an HTTPS Certificate')

			CONST = HTTPSCONST()
			
			# Check if uploaded certificate exists
			if os.path.exists(os.path.join(temporarydirectory,certificatefilename)):
				logger.info('Found uploaded ' + certificatefilename + ' under ' + temporarydirectory)
				global uploaded_https_certificate_found
				uploaded_https_certificate_found = True
			else:
				logger.error('Certificate file ' + certificatefilename + ' NOT found under ' + temporarydirectory)
				failed_exit()

			# move uploaded files to SSL certificate location
			if uploaded_https_certificate_found:
				verify_file(os.path.join(temporarydirectory,certificatefilename), deploytype, temporarydirectory)
				verify_valid_pem_certificate(os.path.join(temporarydirectory,certificatefilename))
				# cleanup and create new backup directory and move original files
				create_directory(CONST.SSL_CERTIFICATE_BACKUP_DIR)
				if os.path.exists(CONST.SSL_CERTIFICATE_FILE):
					move_file(CONST.SSL_CERTIFICATE_FILE, CONST.SSL_CERTIFICATE_BACKUP_DIR + CONST.SSL_CERTIFICATE_FILE_NAME + '.orig')
				if os.path.exists(CONST.SSL_CERTIFICATE_KEY_FILE):
					move_file(CONST.SSL_CERTIFICATE_KEY_FILE, CONST.SSL_CERTIFICATE_BACKUP_DIR + CONST.SSL_CERTIFICATE_KEY_FILE_NAME + '.orig')
				if os.path.exists(CONST.SSL_CERTIFICATE_DIR + certificatefilename):
					move_file(CONST.SSL_CERTIFICATE_DIR + certificatefilename, CONST.SSL_CERTIFICATE_BACKUP_DIR + certificatefilename + '.orig')

				if certificatefilename == CONST.SSL_CERTIFICATE_FILE_NAME:
					logger.info('Uploaded certificate file name is ' + certificatefilename)
				else:
					logger.info('Uploaded certificate file name is ' + certificatefilename + ' ,needs to be changed.')
					global change_https_certificate_file_name
					change_https_certificate_file_name = True

				copy_file(os.path.join(temporarydirectory,certificatefilename), CONST.SSL_CERTIFICATE_DIR)
				if change_https_certificate_file_name:
					logger.info('Changing certificate file name to ' + CONST.SSL_CERTIFICATE_FILE_NAME)
					move_file(CONST.SSL_CERTIFICATE_DIR + certificatefilename, CONST.SSL_CERTIFICATE_FILE)
    
				# extract private key to seperate file
				extract_private_key()
				if (os.path.exists(CONST.SSL_CERTIFICATE_FILE) and os.path.exists(CONST.SSL_CERTIFICATE_KEY_FILE)):
					logger.info('Found ' + CONST.SSL_CERTIFICATE_FILE + ' and ' + CONST.SSL_CERTIFICATE_KEY_FILE + ' files.')
					#delete_directory(temporarydirectory)
					clean_exit()
				else:
					logger.error('Unable to find ' + CONST.SSL_CERTIFICATE_FILE + ' and ' + CONST.SSL_CERTIFICATE_KEY_FILE + ' files.')
					#delete_directory(temporarydirectory)
					failed_exit()

		elif deploytype == "ldap":

			logger.info('Processing an LDAP Certificate')

			CONST = LDAPCONST()

			# Check if uploaded certificate exists
			if os.path.exists(os.path.join(temporarydirectory,certificatefilename)):
				logger.info('Found uploaded ' + certificatefilename + ' under ' + temporarydirectory)
				global uploaded_ldap_certificate_found
				uploaded_ldap_certificate_found = True
			else:
				logger.error('Certificate file ' + certificatefilename + ' NOT found under ' + temporarydirectory)
				failed_exit

			# Handle LDAP certificate
			if uploaded_ldap_certificate_found:
				verify_file(os.path.join(temporarydirectory,certificatefilename), deploytype, temporarydirectory)
				process_ldap_cert(temporarydirectory, certificatefilename)

		else:
			logger.info('Unknown mode, please check usage')    
			failed_exit()

	else:
	    logger.info('Unknown mode, please check usage')    
	    failed_exit()

if __name__ == "__main__":
	main(sys.argv)
