# ==============================================================================
# 
# Copyright (C) 2018 Catalogic Software, Inc., All Rights Reserved
#
# Log parser for Spectrum Protect Plus (SPP) Virgo logs. This tool is 
# responsible for consuming a set of input log files (any combination of
# text-based log files, such as .log files, and Gzipped-compressed log files,
# such as .gz files), filtering out undesired log entries, and writing the
# retained log entries to a destination file.
#
# This tool parses the input set of log files and allows for filters to be
# registered that accept or reject log entries (single lines in the input log
# files) to be written to the output log file.
#
# Implementation Details:
# This parser uses a binary search algorithm to find the first line to read (the
# first line that is equal to or after the provided start time), treating all 
# provided log file as a single file and starting at the middle of this composed
# log file. Note that the composed log file is not completely read into memory, 
# but instead, each log file that makes up this composed log file is read into 
# memory when needed (i.e. at most, one log file is read into memory at a time).
#
# :author: Justin Albano <jalbano@catalogicsoftware.com>
#
# ==============================================================================

import abc
import argparse
import re
import gzip
import os
import dateutil.parser
import sys
import datetime
import struct
import math

from os import listdir
from os.path import isfile, join
from datetime import tzinfo


# ------------------------------------------------------------------------------
# Exceptions
# ------------------------------------------------------------------------------


class LineNotFoundException(Exception):
    pass


# ------------------------------------------------------------------------------
# Filters
# ------------------------------------------------------------------------------


class LogLineFilter(object):
    """
    Interface representing a log line filter. This filter accepts a log line 
    (the log line should be written to the output log file) if the test method 
    returns True; this filter rejects a log line (the log line should not be 
    written to the output log file) if the test method returns False.

    :author: Justin Albano <jalbano@catalogicsoftware.com>
    """
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def test(self, line, context):
        raise NotImplementedError()


class LogLineFilterContext(object):
    """
    Context that represents the current session in which an output log file is
    being written.

    :author: Justin Albano <jalbano@catalogicsoftware.com>
    """

    def __init__(self):
        self.previous_line = None
        self.previous_line_was_included = False


class CompositeFilter(LogLineFilter):
    """
    A composite filter that tests each of the sub-filters in sequential order. 
    If any sub-filter is found that does not accept a specified line, the line 
    is rejected (i.e. uses a logical AND to test each sub-filter).

    :author: Justin Albano <jalbano@catalogicsoftware.com>
    """

    def __init__(self, filters=[]):
        self._filters = filters

    def add_filter(self, log_filter):
        self._filters.append(log_filter)

    def remove_filter(self, log_filter):
        self._filters.remove(log_filter)

    def test(self, line, context):

        for log_filter in self._filters:
            if log_filter.test(line, context) is False:
                return False

        return True


class JobSessionFilter(LogLineFilter):
    """
    Filter that accepts log lines with a job session ID that any matches the
    supplied job session IDs. If a log line does not have a job session ID, it
    is accepted by default (this behavior can be changed using the constructor
    arguments).

    :author: Justin Albano <jalbano@catalogicsoftware.com>
    """

    pattern = re.compile(r'^(\d+)\s+.+')

    def __init__(self, job_session_ids=[], accept_lines_with_missing_job_session_id=True):
        self._job_session_ids = job_session_ids
        self._accept_lines_with_missing_job_session_id = accept_lines_with_missing_job_session_id

    def test(self, line, context):
        
        if not line.has_timestamp():
            return True
        else:
            extracted_job_session_id = JobSessionFilter.extract_job_session_id(line.message)

            if extracted_job_session_id is None and self._accept_lines_with_missing_job_session_id:
                return True
            else:
                return extracted_job_session_id in self._job_session_ids

    @staticmethod
    def extract_job_session_id(message):
        match = JobSessionFilter.pattern.search(message)

        if match is None:
            return None
        else:
            return match.group(1)

class AcceptNonTimestampLinesIfPreviousLineWasIncludedFilter(LogLineFilter):

    def test(self, line, context):
        return line.has_timestamp() or (not line.has_timestamp() and context.previous_line_was_included)


# ------------------------------------------------------------------------------
# Log Parsing
# ------------------------------------------------------------------------------


class LogLine(object):
    """
    A single line in a log file.

    :author: Justin Albano <jalbano@catalogicsoftware.com>
    """

    pattern = re.compile(r'(\[(\S+)\]\s+([a-zA-Z]+)\s+(\S+)\s+(\S+)\s+)?(.+)$')

    def __init__(self, timestamp, raw_timestamp, level, thread, component, message, raw_line):
        self.timestamp = timestamp
        self._raw_timestamp = raw_timestamp
        self.level = level
        self.thread = thread
        self.component = component
        self.message = message
        self._raw_line = raw_line

    def has_timestamp(self):
        return self._raw_timestamp is not None

    @staticmethod
    def from_line(line, skip_parse_timestamp=False):
        """
        Parses a line of log text into a LogLine object. The slf4j format for
        log entries are used for parsing a textual line of a log. This format
        generally resembles the following:

            [<timestamp>] <level> <thread> <component> <message>

        For example:

            [2018-07-26T17:17:08.337Z] INFO  pool-6-thread-1
            \tcom.syncsort.dp.xsb.sessionmanager.impl.SessionManagerImpl
            \tSessionManager: Destroying session with ID: 123456

        :param line:
            The textual log line

        :returns:
            A LogLine object that represented the parsed textual log line.
        """
        stripped_line = line.rstrip()
        match = LogLine.pattern.search(stripped_line)

        raw_timestamp = match.group(2)
        timestamp = None

        if not skip_parse_timestamp:

            if raw_timestamp is not None:
                # Parse the timestamp if one is present
                timestamp = dateutil.parser.parse(match.group(2))

        return LogLine(
            timestamp,
            raw_timestamp,
            match.group(3),
            match.group(4),
            match.group(5),
            match.group(6),
            stripped_line
        )

    def as_dict(self):
        return {
            'timestamp': self.timestamp,
            'level': self.level,
            'thread': self.thread,
            'component': self.component,
            'message': self.message
        }

    @property
    def bytes(self):
        # +1 for new line character
        return len(self._raw_line) + len(os.sep)

    def __str__(self):
        return self._raw_line
    

class TimeBounds(object):

    def __init__(self, start, end):
        self._start = start
        self._end = end

    @property
    def start(self):
        return self._start

    @property
    def end(self):
        return self._end


class Log(object):
    """
    An abstraction of a log responsible for providing an iterable representation
    of a log, where each iterated element is a LogLine object.

    :author: Justin Albano <jalbano@catalogicsoftware.com>
    """
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def next_log_line(self, skip_parse_timestamp=False):
        raise NotImplementedError()

    @abc.abstractmethod
    def seek(self, offset):
        raise NotImplementedError()

    def reset_seek_index(self):
        self.seek(0)

    @abc.abstractmethod
    def get_current_seek_index(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def get_size_in_bytes(self):
        raise NotImplementedError()

    def __iter__(self):
        return self

    def next(self):
        return self.next_log_line()

    @abc.abstractmethod
    def get_latest_possible_timestamp(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def get_earliest_possible_timestamp(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def get_start_seek_index_upper_bound(self):
        raise NotImplementedError()

    @abc.abstractmethod
    def get_end_seek_index_lower_bound(self):
        raise NotImplementedError()

    def write_to_file(self, path, bounds, other_filter=None):
        
        context = LogLineFilterContext()
        output_file = open(path, 'w')

        results = LogResults()

        stop_position = None
        current_position = None

        if bounds.end < self.get_earliest_possible_timestamp():
            print('     [INFO] Provided end is before start of logs')
            return results

        try:
            # Find the end line first so that once the start line is found, the
            # current seek index is position right after the start line
            stop_line = self._find_last_accepted_line(bounds.end)

            if stop_line.is_after:
                # The stop line is the first line after the last acceptable
                # line, so the previous line should be taken
                stop_position = stop_line.previous_end_of_line_position
            else:
                # The stop line is the last line that should be accepted
                stop_position = stop_line.end_of_line_position
            
        except LineNotFoundException as e:
            # Last line is past the end of the log
            stop_position = self.get_size_in_bytes() - 1

        try:
            first_found_line = self._find_first_accepted_line(bounds.start)
            start_line = first_found_line.line
            current_position = first_found_line.end_of_line_position

            if self._line_should_be_included(start_line, context):
                self._record_line(output_file, start_line, context, results)
                print('     Found first accepted line')

        except LineNotFoundException as e:
            print('     [INFO] Provided start timestamp is after the last timestamp found in the logs (no entries to write to output file)')
            return results

        while self.get_current_seek_index() < stop_position:

            line = self.next_log_line(skip_parse_timestamp=True)
            
            if self._line_should_be_included(line, context):
                self._record_line(output_file, line, context, results)
            else:
                context.previous_line_was_included = False

            context.previous_line = line

        return results

    def _line_should_be_included(self, line, context):
        return log_filter is None or log_filter.test(line, context) is True

    def _record_line(self, output_file, line, context, results):
        output_file.write('{}{}'.format(str(line), os.linesep))
        context.previous_line_was_included = True
        results.increment_lines_included()

    def _find_first_accepted_line(self, desired_timestamp):
        return self._find_first_line_at_or_after_timestamp(desired_timestamp, 0, self.get_start_seek_index_upper_bound())

    def _find_last_accepted_line(self, desired_timestamp):
        return self._find_first_line_at_or_after_timestamp(desired_timestamp, self.get_end_seek_index_lower_bound(), self.get_size_in_bytes() - 1)

    def _find_first_line_at_or_after_timestamp(self, desired_timestamp, initial_lower_bound, initial_upper_bound):
        """
        Finds the first acceptable line. An acceptable line is the first line
        that matches the supplied timestamp.

        Implementation Notes:
        This algorithm uses a binary search to find the first acceptable log 
        entry. This binary search algorithm is modified to handle the 
        intricacies of searching by timestamp:

            1.  The search criteria matches the log entry whose timestamp either
                equals the supplied timestamp or is the closest over
            2.  There may be multiple log entries with acceptable supplied dates 
                (and the first should be returned)

        To solve these issues, a binary search is used to find the any entry 
        that matches the supplied timestamp, minus 1 mircrosecond (us)--the 
        smallest discrete increment of time available. This leaves 3 scenarios 
        for the last line checked at the termination of the binary search:

            1.  It is more than 1 us before the supplied timestamp
            2.  It is exactly 1 us before the supplied timestamp
            3.  It is more than 1 us after the supplied timestamp (the last 
                checked line is the next closest after the supplied timestamp, 
                minus 1 us)

        Once this search is complete, each line is read sequentially until an 
        acceptable line is found. In case (1), it is known that there is an 
        acceptable line in the log, since there are entries after the supplied 
        time (hence the search finding a "low end" result); in case (2), 
        sequentially reading the log lines will eventually find an acceptable 
        line or will reach the end of the file, in which case, there is no 
        acceptable line (i.e. the supplied supplied time is after the last entry
        in the log); in case (3), the current line may be the acceptable line, 
        or else the same situation as (2) arises.
        """

        setpoint = desired_timestamp - datetime.timedelta(microseconds=1)

        lower_bound = initial_lower_bound
        middle = 0 
        upper_bound = initial_upper_bound

        while lower_bound <= upper_bound:

            middle = int(math.floor((upper_bound + lower_bound) / 2))
            line = self._get_closest_line_after(middle)

            if line is None:
                raise LineNotFoundException()

            if line.timestamp < setpoint:
                lower_bound = middle + 1

            elif line.timestamp > setpoint:
                upper_bound = middle - 1

            else:
                break

        # Seek to the last checked line: At this point, it is only possible to be. Therefore, we 
        # can iterate forward until the a line is found that is within the 
        # bounds of the supplied time
        self.seek(middle)

        end_of_last_line = middle

        for line in self:
            # Iterate until an accepted first line is found

            if line.timestamp is not None:
                if line.timestamp > desired_timestamp:
                    return FoundLine(line, self.get_current_seek_index(), end_of_last_line, True)
                elif line.timestamp == desired_timestamp:
                    return FoundLine(line, self.get_current_seek_index(), end_of_last_line, False)

            end_of_last_line = self.get_current_seek_index()

        raise LineNotFoundException()


    def _get_closest_line_after(self, offset):
        self.seek(offset)
        
        for line in self:
            if line.timestamp is not None:
                return line


class FoundLine(object):

    def __init__(self, line, end_of_line_position, previous_end_of_line_position, is_after):
        self.line = line
        self.previous_end_of_line_position = previous_end_of_line_position
        self.end_of_line_position = end_of_line_position
        self.is_after = is_after

    def __str__(self):
        return '<position: {}, is after: {}, line: {}>'.format(self.end_of_line_position, self.is_after, self.line)


class LogResults(object):
    """
    Representation of the results of writing log entries to an output file.

    :author: Justin Albano <jalbano@catalogicsoftware.com>
    """

    def __init__(self):
        self._lines_included = 0

    def increment_lines_included(self):
        self._lines_included += 1

    @property
    def lines_included(self):
        return self._lines_included

        

class LogFile(Log):
    """
    A wrapper for an underlying log file on the file system.

    :author: Justin Albano <jalbano@catalogicsoftware.com>
    """
    __metaclass__ = abc.ABCMeta

    def __init__(self, path, earliest_possible_timestamp=None, use_modified_date_for_last_possible_timestamp=False):
        self._file_handle = None
        self._path = path
        self._earliest_possible_timestamp = earliest_possible_timestamp
        self._use_modified_date_for_last_possible_timestamp = use_modified_date_for_last_possible_timestamp
        self._current_seek_index = 0

    @property
    def path(self):
        return self._path

    def next_log_line(self, skip_parse_timestamp=False):

        self._open_file_handle()

        try:
            while True:
                # Read until a non-blank (non-whitespace) line is found
                line = next(self._file_handle)

                self._current_seek_index += len(line)

                if line.strip():
                    # A non-blank (non-whitespace) line was found
                    return LogLine.from_line(line, skip_parse_timestamp)

        except StopIteration as e:
            # Reached the end of the file
            self._file_handle.close()
            self._file_handle = None
            raise e

    def _open_file_handle(self):

        if self._is_not_opened():
            self._file_handle = self.get_file_handle(self._path)

        return self._file_handle

    def _is_not_opened(self):
        return self._file_handle is None

    def seek(self, offset):
        self._open_file_handle()
        self._current_seek_index = offset
        self._file_handle.seek(offset)

    def get_current_seek_index(self):
        self._open_file_handle()
        return self._current_seek_index

    @abc.abstractmethod
    def get_file_handle(self, path):
        raise NotImplementedError()

    def get_start_seek_index_upper_bound(self):
        return self.get_size_in_bytes() - 1

    def get_end_seek_index_lower_bound(self):
        return 0

    def get_latest_possible_timestamp(self):
        mtime = os.path.getmtime(self.path)
        return datetime.datetime.utcfromtimestamp(mtime).replace(tzinfo=Utc())

    def _get_first_line_with_timestamp(self):

        self._open_file_handle()

        first_line = None

        for line in self:
            if line.timestamp is not None:
                return line

    def get_earliest_possible_timestamp(self):
        
        if self._earliest_possible_timestamp is None:
            return self._get_first_line_with_timestamp().timestamp
        else:
            return self._earliest_possible_timestamp

    def __str__(self):
        return os.path.basename(self.path)


class Utc(tzinfo):

  def utcoffset(self, dt):
    return datetime.timedelta(hours=0)

  def tzname(self, dt):
    return "UTC"

  def dst(self, dt):
    return datetime.timedelta(hours=0)


class SizeCachedLogFile(LogFile):

    def __init__(self, path, earliest_possible_timestamp=None, use_modified_date_for_last_possible_timestamp=False):
        LogFile.__init__(self, path, earliest_possible_timestamp, use_modified_date_for_last_possible_timestamp)
        self._cached_file_size = None

    def get_size_in_bytes(self):

        if self._cached_file_size is None:
            self._cached_file_size = self._get_size_in_bytes_first_time()

        return self._cached_file_size

    @abc.abstractmethod
    def _get_size_in_bytes_first_time(self):
        raise NotImplementedError()

class TextLogFile(SizeCachedLogFile):
    """
    A text-based log file.

    :author: Justin Albano <jalbano@catalogicsoftware.com>
    """

    def __init__(self, path, earliest_possible_timestamp=None, use_modified_date_for_last_possible_timestamp=False):
        SizeCachedLogFile.__init__(self, path, earliest_possible_timestamp, use_modified_date_for_last_possible_timestamp)

    def get_file_handle(self, path):
        return open(path)

    def _get_size_in_bytes_first_time(self):
        return os.stat(self._path).st_size

    def as_formatted_string(self):
        return 'TextLogFile<{}>'.format(self.path)


class GzippedLogFile(SizeCachedLogFile):
    """
    A text-based log file that is compressed using Gzip.

    :author: Justin Albano <jalbano@catalogicsoftware.com>
    """

    
    def __init__(self, path, earliest_possible_timestamp=None, use_modified_date_for_last_possible_timestamp=False):
        SizeCachedLogFile.__init__(self, path, earliest_possible_timestamp, use_modified_date_for_last_possible_timestamp)

    def get_file_handle(self, path):
        return gzip.open(path, 'rb')

    def _get_size_in_bytes_first_time(self):
        with open(self._path, 'rb') as f:
            f.seek(-4, 2)
            return struct.unpack('I', f.read(4))[0]

    def as_formatted_string(self):
        return 'GizppedLogFile<{}>'.format(self.path)


class CompositeLog(Log):
    """
    An abstraction that is responsible for combining multiples logs into a 
    single, continuous log. In general, this log combines a set of sub-logs into
    a single, continuous log. Note that the order in which the sub-logs are 
    added is the order in which they are read. The functionality of this log is 
    based on three distinct cases:

        1. If no sub-logs are provided, this abstraction acts 
           as an empty log and immediately throws a StopIteration error. 
        2. If this log contains a single log, it acts identical to any other 
           log, reading lines from sub-log and throwing a StopIteration error 
           when the end of the log is reached. 
        3. If this log contains two or more sub-logs, this log reads completely
           from the first sub-log. Once the end of the first sub-log is reached,
           the second sub-log is read from (starting on the first line of the 
           second sub-log). This process continuous until the final sub-log is 
           reached. Once the end of the final sub-log is reached, a 
           StopIteration error is thrown.

    :author: Justin Albano <jalbano@catalogicsoftware.com>
    """

    def __init__(self, is_run_from_spp=False):
        self._logs = []
        self._current_log = None
        self._is_run_from_spp = is_run_from_spp

    def add_log(self, log):
        self._logs.append(log)

    def remove_log(self, log):
        self._logs.remove(log)

    @property    
    def logs(self):
        return self._logs

    def next_log_line(self, skip_parse_timestamp=False):

        if self._has_no_logs():
            # No sub-logs to read from
            raise StopIteration()
        elif self._has_not_been_initialized():
            # Initialize the sub-logs
            self._initialize()

        if self._is_reading_from_last_sub_log():

            try:
                # Read from last sub-log
                return self._current_log.next_log_line(skip_parse_timestamp)
            except StopIteration as e:
                # Completed reading from all sub-logs
                self._on_successfully_finished_reading_sub_log(self._current_log)
                raise e
            except:
                # An error occurred while reading from the last sub-log
                self._on_failed_to_read_sub_log(self._current_log)
                raise
        else:

            try:
                # Read from an intermediary sub-log (0..n-1 of n sub-logs)
                return self._current_log.next_log_line(skip_parse_timestamp)
            except StopIteration as e:
                # Reached the end of an intermediary: Rotate to next sub-log
                self._on_successfully_finished_reading_sub_log(self._current_log)
                self._rotate_to_next_sub_log()
                return self._current_log.next_log_line(skip_parse_timestamp)
            except:
                # An error occurred while reading from an intermediary sub-log
                self._on_failed_to_read_sub_log(self._current_log)
                raise

    def _has_no_logs(self):
        return len(self._logs) < 1

    def _has_not_been_initialized(self):
        return self._current_log is None

    def _initialize(self):
        self._current_log = self._logs[0]
        self._on_sub_log_started(self._current_log)

    def _is_reading_from_last_sub_log(self):
        return self._current_log is self._logs[-1]

    def _rotate_to_next_sub_log(self):
        current_sub_log_index = self._logs.index(self._current_log)
        self._current_log = self._logs[current_sub_log_index + 1]
        self._current_log.reset_seek_index()
        self._on_sub_log_started(self._current_log)

    def _on_sub_log_started(self, log):
        pass

    def _on_successfully_finished_reading_sub_log(self, log):
        pass

    def _on_failed_to_read_sub_log(self, log):
        pass

    def __str__(self):
        return str([str(log) for log in self.logs])

    def seek(self, offset):
        
        if self._has_no_logs():
            raise IndexError()

        file_sizes = self._get_log_sizes_in_bytes()
        index_of_log_containing_desired_offset = self._index_of_log_containing_byte_index(offset)

        self._current_log = self._logs[index_of_log_containing_desired_offset]
        bytes_in_previous_log_files = sum(file_sizes[:index_of_log_containing_desired_offset])
        offset_into_file = offset - bytes_in_previous_log_files
       
        self._current_log.seek(offset_into_file)

    def get_current_seek_index(self):

        if self._has_no_logs():
            # No sub-logs to read from
            raise StopIteration()
        elif self._has_not_been_initialized():
            # Initialize the sub-logs
            self._initialize()

        this_index = self._logs.index(self._current_log)
        file_sizes_up_to_this = map(lambda l:l.get_size_in_bytes(), self._logs[:this_index])
        return self._current_log.get_current_seek_index() + sum(file_sizes_up_to_this)


    def _index_of_log_containing_byte_index(self, byte_index):
        
        file_sizes = self._get_log_sizes_in_bytes()

        this_log_starting_byte_index = 0
        this_log_ending_byte_index = 0
        current_log_index = 0

        for file_size in file_sizes:

            this_log_ending_byte_index += file_size

            if byte_index >= this_log_starting_byte_index and byte_index < this_log_ending_byte_index:
                return current_log_index
            else:
                this_log_starting_byte_index += file_size
                current_log_index += 1

        raise IndexError()

    def _get_log_sizes_in_bytes(self):
        return map(lambda f:f.get_size_in_bytes(), self._logs)

    def get_size_in_bytes(self):
        return sum(self._get_log_sizes_in_bytes())

    def get_latest_possible_timestamp(self):
        timestamps = map(lambda l:l.get_latest_possible_timestamp(), self._logs)
        return max(timestamps)

    def get_earliest_possible_timestamp(self):
        timestamps = map(lambda l:l.get_earliest_possible_timestamp(), self._logs)
        return min(timestamps)

    def get_start_seek_index_upper_bound(self):
        
        if len(self._logs) == 0:
            return 0
        elif self._is_run_from_spp:
            return self._logs[0].get_start_seek_index_upper_bound()
        else:
            return self.get_size_in_bytes() - 1

    def get_end_seek_index_lower_bound(self):

        if len(self._logs) == 0:
            return 0
        elif self._is_run_from_spp:
            all_but_last_file_sizes = map(lambda l:l.get_size_in_bytes(), self._logs[:-1])
            return sum(all_but_last_file_sizes) - 1
        else:
            return 0


class LogDirectory(CompositeLog):
    """
    An abstraction that is responsible for providing a composite log 
    representation of a directory containing .log and .log.gz log files. The
    order of the logs in the composite representation is based on the name of
    the file, where logs with the format log_<num> are consumed in order of
    decreasing <num> and the file log.log is consumed last. 

    For example, a directory containing the following logs

        log_1.log log_2.log.gz log.log

    is composed into a CompositeLog that reads the log files in the following
    order:

        1. log_2.log.gz
        2. log_1.log
        3. log.log

    Therefore, the log file with the oldest entries (log_2.log.gz) is read first
    (when the next_log_line method is called) and the log with the newest
    entries (log.log) is read last.

    :author: Justin Albano <jalbano@catalogicsoftware.com>
    """

    def __init__(self, directory, start_time, end_time, is_run_from_spp=False):
        CompositeLog.__init__(self, is_run_from_spp)

        all_files = [f for f in listdir(directory) if isfile(join(directory, f))]
        log_file_names = filter(LogDirectory.find_log_files, all_files)
        log_file_names.sort(reverse=True)

        earliest_possible_timestamp = None

        for log_file_name in log_file_names:
            log_file_path = join(directory, log_file_name)

            if log_file_path.endswith('.log.gz'):
                log = GzippedLogFile(log_file_path, earliest_possible_timestamp, is_run_from_spp)
            else:
                log = TextLogFile(log_file_path, earliest_possible_timestamp, is_run_from_spp)

            if is_run_from_spp:

                if log.get_latest_possible_timestamp() >= start_time and log.get_earliest_possible_timestamp() <= end_time:
                    self.add_log(log)
                else:
                    print('     [INFO] Ignoring {} because the last possible timestamp it contains is before the desired start time'.format(str(log)))
                
                earliest_possible_timestamp = log.get_latest_possible_timestamp()

            else:
                self.add_log(log)
               

    @staticmethod
    def find_log_files(file):
        return file.endswith('.log') or file.endswith('.log.gz')


# ------------------------------------------------------------------------------
# Application
# ------------------------------------------------------------------------------


if __name__ == "__main__":

    parser = argparse.ArgumentParser(description='Spectrum Protect Plus (SPP) Virgo Log Parser')
    parser.add_argument('-i', '--input', action="store", default='/opt/virgo/serviceability/logs/')
    parser.add_argument('-o', '--output', action="store", required=True, help='The path to the output file to which the parsed log should be dumped.')
    parser.add_argument('-s', '--start', action="store", required=True, help='The start timestamp. All log messages equal to or after this timestamp, inclusive, (up to the end tiemstamp) are included.')
    parser.add_argument('-e', '--end', action="store", required=True, help='The end timestamp. All log messages before this timestamp, inclusive, (after the start timestamp) are included.')
    parser.add_argument('--standalone', dest='is_not_running_from_spp', default=False, action="store_true", help='Disables optimization assumptions that are valid only when run from SPP instance.')
    parser.add_argument('-j', '--job-session-ids', dest='job_session_ids', action="store", nargs='+', help='The list of all job session IDs whose messages should be included. If a log entry has no associated job session ID, it will be included (unless another filter excludes it).')
    args = parser.parse_args()

    start = dateutil.parser.parse(args.start)
    end = dateutil.parser.parse(args.end)

    if start > end:
        print('Start time must be before or equal to end time')
        exit(1)

    log_filter = CompositeFilter()

    timestamp_log_filter = TimeBounds(start, end)

    log_filter.add_filter(AcceptNonTimestampLinesIfPreviousLineWasIncludedFilter())

    if args.job_session_ids is not None:
        job_session_filter = JobSessionFilter(args.job_session_ids)
        log_filter.add_filter(job_session_filter)

    print ''
    print '============================================================================================='
    print ' SPP Virgo Log Parser'
    print '============================================================================================='
    print ''
    print ' Configuration'
    print ' ============='
    print ''
    print '   Input directory:       {}'.format(args.input)
    print '   Output file:           {}'.format(args.output)
    print '   Starting timestamp:    {}'.format(args.start)
    print '   Ending timestamp:      {}'.format(args.end)
    print ''
    print ''
    print ' Parsing Log'
    print ' ==========='
    print ''
    print '   Parsing log files...'

    tick = datetime.datetime.now()
    log = LogDirectory(args.input, start, end, is_run_from_spp=not args.is_not_running_from_spp)
    results = log.write_to_file(args.output, timestamp_log_filter, log_filter)
    tock = datetime.datetime.now()
    
    print ''
    print ''
    print '============================================================================================='
    print ' Output Results'
    print '============================================================================================='
    print ' Lines included:     {}'.format(results.lines_included)
    print ' Elapsed Time:       {}'.format(str(tock - tick))
    print '============================================================================================='
    print ''
